/*	Allgemeine Info zur Fat !

siehe auch: https://www.pjrc.com/tech/8051/ide/fat32.html
und 			fatgen103.pdf von microsoft

*************************************************************
die fat hat 4 teile: 	1. sektor 0 mit den daten über die fat
								2. den file-allocation-table, der die cluster-chains beinhaltet (eigentlicher fat bereich)
								3. (nur bei fat16 !!!) den root-dir bereich, wo das wurzelverzeichniss liegt  		(rootDir)
								4. den data-dir bereich wo, die daten liegen,oder unterordner,oder bei fat32 auch 
									das root dir. ist in cluster eingeteilt. 														(dataDirSec)
ein directory ist nichts anderes als: der 1. cluster und ein name, einer cluster-chain in der einträge stehen die,
dem namen und dem 1. cluster zugeordnet sind !!

*************************************************************
einziger unterschied zwischen fat16 und fat32 ist die größe der fat eintrage, 2byte bei fat16 bzw. 4byte bei fat32 und der 
umgang mit dem root direktory. das root directory ist bei fat16 nur groß genug für 512 dateien/ordner, weil ein fest 
berechneter/bestimmter berreich für das directory vorgesehen ist. fat32 behandelt auch das root directory wie einen ganz 
normalen ordner (also beliebig viele ordner/dateien möglich)! allerdings steht der start cluster des root directory im sektor 0,
im fat32 spezifischen teil (offset 44, size 4)!

*************************************************************
die fat ist wie folgt aufgebaut:

	(index)		(inhalt)
	cluster	:	folgeCluster
	0			:	1
	1			:	2
	2			:	3
	3			:	5
	4			:	7
	5			:	end
	6			:	end
	7			:	8
	8			:	9
	9			:	10
	10			:	end
	11			:	12
	12			:	end

	4 dateien:								(file.firstCluster)
	1. datei startcluster : 0			
	2. datei startcluster : 4
	3. datei startcluster : 6
	4. datei startcluster : 11
	
	clusterChains der dateien:			(siehe: fat_Cluster_getNextCluster)
	
	1.		0:1:2:3:5
	2.		4:7:8:9:10
	3.		6
	4. 	11:12

wichtig: 

die fat ist seriell durchnummeriert, wobei jeder index (oder stelle) für einen cluster steht. 
der inhalt des clusters ist der folgecluster!
siehe oben, cluster 0 hat den inhalt 1. also ist cluster 0 mit 1 verknüpft.

ein cluster 0 gibt es nicht, damit ist lediglich die stelle 0 in der fat gemeint (cluster beginnen immer mit 2)!! 
*************************************************************
ein direktory eintrag ist wie folgt aufgebaut:		(siehe: fat_loadRowOfSector)
besteht aus 32 bytes

byte 0-10 	=  8.3 datei name (ohne punkt, z.b. DATEI001TXT)																(file.name)
byte 11		=	Attribut (siehe tabelle weiter unten)																			(file.attrib)
byte 20/21	= 	high bytes des 1.clusters der datei																				(file.firstCluster)		
byte 26/27	= 	low bytes des 1.clusters der datei (setzt sich so zusammen: 21:20:27:26=4byte wert)				(file.firstCluster)
byte 28-31	=	größe der datei 4 byte also ergibt sich eine maximale datei größe von 4294967295 bytes/datei (file.size)


attribut bits:

1	Hidden				Should not show in dir listing
2	System				File is operating system
3	Volume ID			Filename is Volume ID
4	Directory			Is a subdirectory (32-byte records)
5	Archive				Has been changed since last backup
6	Ununsed				Should be zero
7 (MSB)	Ununsed		Should be zero
*************************************************************

wichtig:

fat ist : "little endian"
d.h. die bit weritgkeit ist umgedreht. das byte/bit ganz links ist das niedrigste
beispiel:	0x0102030405 als little endian ist als big endian 0x0504030201 !!

*************************************************************
arbeitsweise dieser fat libary:
es wird der sektor 0 der mmc/sd Karte ausgelesen um die wichtigen berreiche und variablen zu bestimmen.
es wird eine daten-chain implementiert, wo die daten von der einen ebende zur anderen bearbeitet werden.
laden des sektors->einzelne reihe analysieren->mehrere reihen analysieren->mehrere sektoren analysieren->clusterChains analysieren


		1. fat_loadSector() prüft ob der grade gepufferte sektor nicht der gleiche wie der übergebene ist und
			läd dann den übergebenen sektor von der karte.

		2. fat_loadRowOfSector() analysiert , davon ausgehend es handelt sich um geladene datei/directory einträge,
			eine reihe der einträge (32 bytes/eintrag). datei name usw. werden auf das file struct geladen !
		
		3. fat_loadFileDataFromCluster() geht mit hilfe von fat_loadRowOfSector() nun durch mehrere reihen des sektors
			und sucht so nach dem übergebenen datei/ordner namen, wenn gefunden, return(0).
			hack: wenn ein cluster aus mehreren sektoren besteht, werden diese auch durchsucht,
			der name fat_loadFileDataFromCluster() passt also nicht ganz..
		
		4. fat_loadFileDataFromDir() sucht mit hilfe von fat_loadFileDataFromCluster() in sektoren nach dem 
			übergebenen datei namen. es wird die clusterChain die zu dem start cluster auf der variablen "dir" gehört 
			durchsucht. ein ordner/dir ist nichts anderes, als eine clusterChain in deren sektoren einträge über
			dateien oder andere ordner stehen. nötig um ein direktory zu durchsuchen ist nur der startcluster,
			da ab da wieder die verkettung der fat greift... 
			z.b. durchsuche sektor 12, nicht gefunden, dann durchsuche sektor 13 usw, bis clusterChain ende, wenn bis
			dahin nicht aufgetaucht ist der ordner/datei nicht vorhanden 

		5. fat_cd() setzt mit hilfe von fat_loadFileDataFromDir() den start cluster des direktory das durch den
			übergebenen damen gesucht wurde auf die variable "dir". "dir" wird selber von fat_loadFileDataFromDir()
			benutzt um dateien/ordner in dem direktory zu suchen.

start immer im root direktory. besonders muss auch darauf eingegangen werden um welchen fat typ es sich handelt,
weil bei fat16 das root dir nicht im datenbereich liegt !

*/


// optimierung: in file.c die variable seek registrieren in register

//***************************************************************************************************************
unsigned char 		fat_initfat(void);									// initalisierung (durchsucht MBR oder nicht)
unsigned char 		fat_loadFatData(unsigned long int);					// läd fat daten 

unsigned long int 	fat_clustToSec(unsigned long int);					// rechnet cluster zu 1. sektor des clusters um
unsigned char 		fat_writeSector(unsigned long int sec);				// schreibt sektor auf karte
void 				fat_markSector00(void);								// markiert puffer:sector mit 0x00

unsigned long int 	fat_getNextCluster(unsigned long int oneCluster);		// fat auf nächsten, verketteten cluster durchsuchen 
unsigned long int 	fat_getFreeCluster(unsigned long int offsetCluster);	// sucht leeren fat cluster aus der fat
unsigned char		fat_setCluster(unsigned long int cluster, unsigned long int content); // setzt cluster inhalt in der fat
unsigned char 		fat_delClusterChain(unsigned long int startCluster);	// löscht cluster-chain in der fat


// fat_getFreeRowOfCluster -> fat_getFreeRowOfDir -> fat_makeRowDataEntry -> fat_makeFileEntry -> fat_writeSector  "eintrag gemacht !!"
unsigned char     	fat_getFreeRowOfCluster(unsigned long secStart);	// durchsucht cluster nach freiem eintrag
unsigned char 		fat_getFreeRowOfDir(unsigned long int dir);			// durchsucht directory nach feiem eintrag
unsigned char 		fat_makeRowDataEntry(unsigned int row,unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length);
unsigned char 		fat_makeFileEntry(unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length);

// fat_loadSector -> fat_loadRowOfSector -> fat_loadFileDataFromCluster -> fat_loadFileDataFromDir -> fat_loadFileDataFromDir -> fat_cd   "daten chain"
unsigned char 		fat_loadSector(unsigned long int sec);				// läd übergebenen absoluten sektor
unsigned char 		fat_loadRowOfSector(unsigned int);					// läd reihe des geladen sektors auf struct:file
unsigned char 		fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name[]);	// durchsucht die reihen des geladenen sektors 
unsigned char 		fat_loadFileDataFromDir(unsigned char name[]);		// durchsucht das aktuelle directory 
unsigned char		fat_cd(unsigned char *);							// wechselt directory (start im rootDir)

//***************************************************************************************************************

#define superfloppy 1	// wenn 0 ist partitioniert, wenn 1 ist unpartitioniert (superfloppy) !!

unsigned char 	sector[512];				// sektor puffer
unsigned char	*bytesOfSec=&sector[0];		// byte zeiger auf puffer:sector
unsigned char 	bufferDirty=0;					// zeigt an ob buffer dirty (wird von fat_setCluster gesetzt und von fat_loadSector gelesen)


struct Fat{									// fat daten (1.cluster, root-dir, dir usw.)
	unsigned char 		secPerClust;		// anzahl der sektoren pro cluster
	unsigned long int 	currentSectorNr; 	// aktuell geladener Sektor (in sector)  //beschleunigt wenn z.b 2* 512 byte puffer vorhanden, oder bei fat operationen im gleichen sektor		
	unsigned long int 	dir; 		  		// Direktory zeiger rootDir=='0' sonst(1.Cluster des dir; start auf root)	
	unsigned long int  	rootDir;			// Sektor(f16)/Cluster(f32) nr root directory
	unsigned long int  	dataDirSec;			// Sektor nr data area 
	unsigned long int  	fatSec;	 			// Sektor nr fat area
	unsigned int  		bytsPerSec;			// bytes per sector	
	unsigned char  		fatType;			// fat16 oder fat32 (16 oder 32)
	
	} fat;


struct File{								// datei infos	(32 bytes lang- 16 bei 512 bytesPerSec)												
	unsigned char 		name[13];			// 0,10			datei Name.ext (8.3 = max 11)
	unsigned char		attrib;				// 11,1			datei Attribut: 8=value name, 32=datei, 16=Verzeichniss, 15=linux kleingeschrieben eintrag
	unsigned long int 	firstCluster;		// 20,2 /26,2	datei 1.cluster hi,low(möglicherweise der einzige)	(4-byte)
	unsigned long int	length;				// 28,4			datei Länge (4-byte)				
	unsigned char 		row;				// reihe im sektor in der die datei infos stehen (reihe 0-15)

	} file;	


//***************************************************************************************************************
// markiert sector komplett mit 0x00 (nur den gepufferten)
// wenn neuer direktory cluster verkettet wird, muss er mit 0x00 initialisiert werden (alle felder)
//***************************************************************************************************************
void fat_markSector00(void){
		
  unsigned int i=511;	
	
  do{
	sector[i]=0x00;		//schreibt puffer sector voll mit 0x00==leer
	}while(--i);		
  sector[0]=0x00;		// noch die stelle 0 überschreiben		
}


//***************************************************************************************************************
// umrechnung cluster auf 1.sektor des clusters (möglicherweise mehrere sektoren/cluster !
//***************************************************************************************************************
unsigned long int fat_clustToSec(unsigned long int clust){

	return fat.dataDirSec+2+((clust-2) * fat.secPerClust);		// errechnet den 1. sektor der sektoren des clusters	
}


//***************************************************************************************************************
// läd sektor:sec auf puffer:sector zum bearbeiten im ram !
// mmc_read_sector setzt currentSectorNr auf richtigen wert (also den sektor der gepuffert ist)
// fat_loadFileDataFromDir <- fat_loadFileDataFromCluster <- fat_loadRowOfSector <- fat_loadSector "daten chain"
//***************************************************************************************************************
unsigned char fat_loadSector(unsigned long int sec){
	
  if(sec!=fat.currentSectorNr){					// nachladen nötig		
	if(bufferDirty==1)fat_writeSector(fat.currentSectorNr);	// puffer diry, also vorher schreiben		  		  
	if(0==mmc_read_sector(sec,sector)){			// neuen sektor laden					
	  fat.currentSectorNr=sec;					// aktualisiert sektor nummer (nummer des gepufferten sektors)
	  return(0);
	  }
	}	
		
  else return(0);								// alles ok, daten sind schon da (sec==fat.currentSectorNr)	

  return(1);									// fehler	
}


//***************************************************************************************************************
// schreibt sektor nummer : sec auf die karte (puffer:sector) !!
// setzt bufferFlag=0 da puffer nicht diry sein kann nach schreiben !
//***************************************************************************************************************
unsigned char fat_writeSector(unsigned long int sec){
	
  if(0==mmc_write_sector(sec,sector)){
	bufferDirty=0;	// buffer kann nicht dirty sein weil grade geschrieben		
	return(0);		// schreiben von sektor puffer											
	}

  return (1);
}






// datei lesen funktionen:

// fat_loadSector -> fat_loadRowOfSector -> fat_loadFileDataFromCluster -> fat_loadFileDataFromDir -> fat_loadFileDataFromDir -> fat_cd   "daten chain"

//***************************************************************************************************************
// geht die reihen, also 32 byte einträge im puffer:sector durch und speichert informationen auf das struct:file in 
// dem alle wichtigen daten stehen, wie: 1.cluster,länge bei dateien, name des eintrags, reihen nummer (im sektor)
//***************************************************************************************************************
unsigned char fat_loadRowOfSector(unsigned int row){

	unsigned char i;												// zähler im sektor(32byte zähler)	
	row=row<<5;														// multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 usw)	
	
	for(i=0;i<32;i++){											// zählt durch "spalte", also einzelne bytes der reihe
	
		if(i<11){
			file.name[i]=sector[row];					// datei name, ersten 10 bytes vom 32 byte eintrag
			}
		
		if(i==11)file.attrib=sector[row];					// datei attribut (1 byte)
					
		if(i==20){								// datei 1.cluster (4 byte)							
			unsigned long int tmp;			// muss um die 4 einzelnen bytes auf file.firstCluster zu bekommen		
			file.firstCluster=0;	

			tmp=sector[row+1];				// hi byte		(21/20)
			file.firstCluster |=(tmp<<24);				
			tmp=sector[row];					// 1 lower byte
			file.firstCluster |=(tmp<<16);	
			tmp=sector[row+7];				// 2 lower byte	(26/27)
			file.firstCluster |=(tmp<<8);
			tmp=sector[row+6];				// lowest byte
			file.firstCluster |=tmp;
			}

		if(i==28){								// datei länge (4 bytes)					
			unsigned long int tmp;			// muss um die 4 einzelnen bytes auf file.length zu bekommen
			file.length=0;						// setzt länge auf null da bitweise oder operationen	

			tmp=sector[row+3];				// hi byte		
			file.length |=(tmp<<24);				
			tmp=sector[row+2];				// 1 lower byte
			file.length |=(tmp<<16);	
			tmp=sector[row+1];				// 2 lower byte
			file.length |=(tmp<<8);
			tmp=sector[row];					// lowest byte
			file.length |=tmp;
			
			return (0);							// alles gelesen
			}		
		row++;			
		}
			
	return(1);				
}


//***************************************************************************************************************
// sucht im sektor:sec (geladen auf puffer:sector) nach dem eintag:name und läd die daten
// auf das struct:file. (siehe auch file_loadRowOfSector), wenn nicht gefunden oder fehler return(1)
//***************************************************************************************************************
unsigned char fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name[]){
	
  unsigned char i;
  unsigned long int j=0;
  do{											// sektoren des clusters prüfen
	if(0==fat_loadSector(sec+j)){				// läd den sektor:sec auf den puffer:sector		
	  for(i=0;i<16;i++){						// zählt zeilennummer (16(zeilen) * 32(spalten) == 512 bytes des sektors)
		fat_loadRowOfSector(i)	;				// zeile 0-15 auf struct:file laden 
		if(0==strncmp((char*)file.name,(char*)name,10)){	// zeile i ist gesuchte
		  file.row=i;										// zeile sichern.
		  return(0);			
		  }
		}
	}
	j++;
	}while(j<fat.secPerClust);					// geht durch sektoren des clusters

	return (1);									// fehler (datei nicht gefunden, oder fehler beim lesen)
}


//***************************************************************************************************************
// wenn dir == "0" dann wird das root direktory durchsucht, wenn nicht wird der ordner cluster-chain gefolgt, um
// die datei zu finden. dir eintrag muss stimmen um die datei zu finden (also vorher ffcd) ! 
//***************************************************************************************************************
unsigned char fat_loadFileDataFromDir(unsigned char name[]){
		
  unsigned long int i=0;			

  if(fat.dir=='0' && fat.fatType==16){									// IM ROOTDIR.	fat16		
	for(i=0;i<(fat.dataDirSec+2-fat.rootDir);i++){							// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).								
	  if(0==fat_loadFileDataFromCluster(fat.rootDir+i,name))return(0);		// sucht die datei, wenn da, läd daten (1.cluster usw)															
	  }
	}		
	
  else {		
	if(fat.dir=='0' && fat.fatType==32)i=fat.rootDir;						// IM ROOTDIR.	fat32			
	else i=fat.dir;																											
	//while((i&0xf0000007 && fat.fatType==32) || (i&0xf0000000&& fat.fatType==32) || (i&0x0000&&fat.fatType==16)){	// prüft ob weitere sektoren zum lesen da sind (fat32||fat16)					=
	while(!((i==0xffffff8&&fat.fatType==32)||(i==0xfffffff&&fat.fatType==32)||(i==0xffff&&fat.fatType==16))){								// prüft ob weitere sektoren zum lesen da sind (fat32||fat16)					=
	  if(0==fat_loadFileDataFromCluster(fat_clustToSec(i),name))return(0);		// lät die daten der datei auf struct:file. datei gefunden (umrechnung auf absoluten sektor)																									
	  i=fat_getNextCluster(i);													// liest nächsten cluster des dir-eintrags (unterverzeichniss größer 16 einträge)								
	  }
	}

  return(1);																	// datei/verzeichniss nicht gefunden
}


//***************************************************************************************************************
// start immer im root Dir
// es MUSS in das direktory gewechselt werden, in dem die datei zum lesen ist (außer root, da startet mann)!
//***************************************************************************************************************
unsigned char fat_cd(unsigned char name[]){

	if(name[0]=='0'){												// ZUM ROOTDIR FAT16/32
		fat.dir='0';																// zeichen root dir 
		return(0);
		}

	if(0==fat_loadFileDataFromDir(name)){			// NICHT ROOTDIR	(fat16/32)
		fat.dir=file.firstCluster;											// zeigt auf 1.cluster des dir	(fat16/32)	
		return(0);
		}

	return(1);														// dir nicht gewechselt (nicht da?!?) !!
}








// datei anlegen funktionen :

//fat_getFreeRowOfCluster -> fat_getFreeRowOfDir -> fat_makeRowDataEntry -> fat_makeFileEntry -> fat_writeSector -> eintrag gemacht !!

// ***************************************************************************************************************
// sucht leeren eintrag im cluster mit dem startsektor:secStart.
// wird dort kein freier eintrag gefunden ist return (1).
// wird ein eintrag gefunden, ist die position der freien reihe auf file.row abzulesen und return (0).
// der sektor mit der freien reihe ist auf dem puffer:sector gepuffert.
// ****************************************************************************************************************
unsigned char fat_getFreeRowOfCluster(unsigned long secStart){
  //printf("\n\rfat_freeRowOfCluster");	
  unsigned int 	i;	
  unsigned char c;	

  for(c=0;c<fat.secPerClust;c++){						// geht die sektoren des clusters durch (möglicherweise auch nur 1 sektor)
	file.row=0;											// neuer sektor, reihen von vorne
	if(0==fat_loadSector(secStart+c)){					// läd sektor auf puffer:buffer.dSector, setzt buffer.currentDatSector
	  for(i=0;i<512;i=i+32){							// zählt durch zeilen (0-15)								
		if(sector[i]==0x00||sector[i]==0xE5)return(0);	// prüft auf freihen eintrag (leer oder gelöscht)
		file.row++;										// zählt reihe hoch (nächste reihe im sektor)
		}
	  }
	}

  return (1);											// nicht gefunden in diesem cluster
}


// ***************************************************************************************************************
// sucht leeren eintrag (zeile) im directory mit dem startcluster:dir.
// geht die cluster chain des direktories durch. dabei werden auch alle sektoren der cluster geprüft.
// wird dort kein freier eintrag gefunden, wird ein neuer leerer cluster gesucht, verkettet und der 
// 1 sektor des freien clusters geladen. die reihe wird auf den ersten eintrag gesetzt, da frei.
// anhand der reihe kann man nun den direktory eintrag vornehmen, und auf die karte schreiben.
// ****************************************************************************************************************
unsigned char fat_getFreeRowOfDir(unsigned long int dir){
  
  unsigned long int start=dir;   
 
  //while( dir!=0xffffff8 && dir!=0xfffffff ||(dir!=0xffff&&fat.fatType==16) ){// solange bis ende cluster chain
  while( !((dir==0xffffff8&&fat.fatType==32)||(dir==0xfffffff&&fat.fatType==32)||(dir==0xffff&&fat.fatType==16)) ){		// solange bis ende cluster chain
  
	if(0==fat_getFreeRowOfCluster(fat_clustToSec(dir)))return(0);	// freien eintrag im cluster gefunden !!
	start=dir;		
	dir=fat_getNextCluster(dir);	
	}										// wenn aus schleife raus, kein freier eintrag da

  dir=fat_getFreeCluster(start);			// muss neuen freien cluster suchen 	(benutzt puffer:sector)
  
  fat_setCluster(start,dir);				// cluster-chain mit neuem verketten	(benutzt puffer:sector)		
  fat_setCluster(dir,0x0fffffff);			// cluster-chain ende markieren	  		(benutzt puffer:sector)
    
  fat_loadSector(fat_clustToSec(dir));		// läd, bzw wenn bufferFlag==1 schreibt vorher, neuen cluster !

  fat_markSector00();						// löschen des sektors (nur im puffer) (benutzt puffer:sector)
  unsigned char i;
  for(i=1;i<fat.secPerClust;i++){			// i=1 weil der 1. eintrag eh noch geschrieben wird... wäre hier doppelt
	fat_writeSector(fat.currentSectorNr+i);	// löschen des cluster (überschreibt mit 0x00), wichtig bei ffls !
	}

  file.row=0;		

  return(0);								// keinen freien platz im dir	
}


//***************************************************************************************************************
// erstellt 32 byte eintrag einer datei, oder verzeichnisses im puffer:sector.
// erstellt eintrag in reihe:row !!  (vorher von fat_getFreeRowOfDir bestimmt)
// muss noch auf die karte geschrieben werden.
//***************************************************************************************************************
unsigned char fat_makeRowDataEntry(unsigned int row,unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length){

	bufferDirty=1;										// puffer beschrieben, also neue daten darin(vor lesen muss geschrieben werden)

	unsigned char i;									//byte zähler in reihe von sektor (32byte eintrag)
	row=row<<5;											//multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 ... zeile 15=480)	

	bytesOfSec=&sector[row];							// init des zeigers
				
	for(i=0;i<32;i++){													

		if(i<11)*bytesOfSec=name[i];					//namen schreiben (format beachten!!) (10 byte)

		if(i==11)*bytesOfSec=attrib;					//attrib schreiben (1 byte)

		if(i>11&&i<20)*bytesOfSec=0x01;					//nulle nicht nötige felder !
		if(i>21&&i<26)*bytesOfSec=0x01;					//nulle nicht nötige felder !

		if(i==20){										// 1. cluster der datei byte 20/21 und 26/27 = 4 bytes
			*bytesOfSec=(cluster&0x00ff0000)>>16;		// 1. low
			bytesOfSec++;	
			i++;
			*bytesOfSec=(cluster&0xff000000)>>24;		// high byte																
			}	
		
		if(i==26){												
			*bytesOfSec=(cluster&0x000000ff);			// low byte								
			bytesOfSec++;	
			i++;
			*bytesOfSec=(cluster&0x0000ff00)>>8;		// 2. low						
			}	
		
		if(i==28){										//datei länge (4 bytes)					
			*bytesOfSec =(length&0x000000ff);			// low			
			bytesOfSec++;	
			*bytesOfSec =(length&0x0000ff00)>>8;		//	1. hi
			bytesOfSec++;	
			*bytesOfSec =(length&0x00ff0000)>>16;		//	2. hi 
			bytesOfSec++;	
			*bytesOfSec =(length&0xff000000)>>24;		//	hi		
			return (0);														// alles ok, eintrag angelegt !!		
			}
		bytesOfSec++;		
		}
		
	return (1);		//fehler (datei nicht gefunden, oder fehler beim lesen)
}


//***************************************************************************************************************
// macht den datei eintrag im jetzigen verzeichniss (dir).
// file.row wird in den unter funktionen gesetzt und hier benutzt.
// WICHTIG !! für fat 16 im root dir muss andere funktion genutze werden, als fat_getFreeRowOfDir
//***************************************************************************************************************
unsigned char fat_makeFileEntry(unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length){
  
  unsigned int i;														// zähler für root dir sektoren fat16
 
  if(fat.dir=='0'&&fat.fatType==32)fat_getFreeRowOfDir(fat.rootDir);	// im root dir (fat32)

  else if(fat.dir=='0' && fat.fatType==16){								// im root dir	(fat16)		
	for(i=0;i<(fat.dataDirSec+2-fat.rootDir);i++){						// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).								
	  if(0==fat_getFreeRowOfCluster(fat.rootDir+i))break;				// geht durch sektoren des root dir.
	  }
	}  

  else fat_getFreeRowOfDir(fat.dir);									// nicht root dir
		
  fat_makeRowDataEntry(file.row,file.name,0x20,file.firstCluster,file.length);	// macht file eintrag im puffer	
  fat_writeSector(fat.currentSectorNr);										  	// schreibt file daten auf karte		  
  
  return(0);
	  
}







// fat funktionen:

//***************************************************************************************************************
// sucht nötige folge Cluster aus der fat ! im 1.Sektor der fat haben 254 cluster(2byte) platz, danach 256
//	erster daten cluster = 2, ende einer cluster chain 0xFFFF, 
// adresse des clusters in der fat, hat als wert, den nächsten cluster. (1:1 gemapt)!
//***************************************************************************************************************
unsigned long int fat_getNextCluster(unsigned long int oneCluster){	

  // FAT 16**************FAT 16	
  if(fat.fatType==16){					
												//	i=oneCluster/256
	unsigned long int i=oneCluster>>8;;			// errechnet den sektor der fat in dem oneCluster ist (rundet immer ab) 
												// j=(oneCluster-256*i)*2 == 2*oneCluster-512*i
	unsigned long int j=(oneCluster<<1)-(i<<9);	// errechnet das low byte von oneCluster in dem ausgelesenen sektor der fat (high und low byte enthalten den folge cluster)
			
	if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden nicht nötig									
	  i=0;  
	  void *vi=&i;								// zeiger auf i
	  bytesOfSec=&sector[j];	  				// zeiger auf puffer
	  *(unsigned char*)vi=*bytesOfSec;			// setzen low byte von i, aus puffer
	  vi++;										// zeigt auf nächst höheres byte in i
	  bytesOfSec++;								// nächst höheres byte in puffer
	  *(unsigned char*)vi=*bytesOfSec;			// setzen von höherem byte in i, aus puffer
	  return i;									// gibt neuen cluster zurück (oder 0xffff)
	  }	
	}

  // FAT 32**************FAT 32	
  if(fat.fatType==32){
												// i=oneCluster/128
	unsigned long int i=oneCluster>>7;			// errechnet den sektor der fat in dem oneCluster ist (rundet immer ab) 
												// j=(oneCluster-128*i)*4 == oneCluster*4-512*i
	unsigned long int j=(oneCluster<<2)-(i<<9);	// errechnet das low byte von oneCluster in dem ausgelesenen sektor der fat (high und low byte enthalten den folge cluster)		
		
	if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden nicht nötig	  	 
	  void *vi=&i;
	  bytesOfSec=&sector[j];
	  *(unsigned char*)vi=*bytesOfSec;
	  vi++;
	  bytesOfSec++;
	  *(unsigned char*)vi=*bytesOfSec;	 
	  vi++;
	  bytesOfSec++;	  
	   *(unsigned char*)vi=*bytesOfSec;
	  vi++;
	  bytesOfSec++;
	  *(unsigned char*)vi=*bytesOfSec;	 
	  return i;
	  }	
  }		

  return(0);									// neuladen des fat sektors, in dem oneCluster ist nötig !!		
}


//***************************************************************************************************************
// sucht freien cluster in der fat ab dem offsetCluster.
//***************************************************************************************************************
unsigned long int fat_getFreeCluster(unsigned long int offsetCluster){
	//printf("\n\rgetFreeCluster");	
		do{			
			offsetCluster++;												// cluster nummer	
			}while(fat_getNextCluster(offsetCluster));			// sucht so lange bis freier cluster gefunden			

	return offsetCluster;							
}


//***************************************************************************************************************
// setzt den cluster inhalt. errechnet den sektor der fat in dem cluster ist, errechnet das low byte von
// cluster und setzt dann byteweise den inhalt:content.
// prüft ob buff dirty (zu setztender cluster nicht in jetzt gepuffertem). 
// wenn flag==1 wird der fat puffer sofort geschrieben (mit den neuen daten).
//***************************************************************************************************************
unsigned char fat_setCluster(unsigned long int cluster, unsigned long int content){  

	// FAT 16**************FAT 16	
	if(fat.fatType==16){					
													//	i=cluster/256
		unsigned long int i=cluster>>8;;			// errechnet den sektor der fat in dem cluster ist (rundet immer ab) 
													// j=(cluster-256*i)*2 == 2*cluster-512*i
		unsigned long int j=(cluster<<1)-(i<<9);	// errechnet das low byte von cluster in dem ausgelesenen sektor der fat 
		
		if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden (fat_loadSector prüft ob schon gepuffert)
			bytesOfSec=&sector[j];					// init des zeigers auf low byte
			*bytesOfSec=(content&0x000000ff);		// low byte					
			bytesOfSec++;			
			*bytesOfSec=(content&0x0000ff00)>>8;	// high byte	
			bufferDirty=1;							// zeigt an, dass im aktuellen sector geschrieben wurde
			return (0); 											
			}	
		}


	// FAT 32**************FAT 32	
	if(fat.fatType==32){
													 // i=cluster/128
		unsigned long int i=cluster>>7;				// errechnet den sektor der fat in dem cluster ist (rundet immer ab) 
													// j=(cluster-128*i)*4 == cluster*4-512*i
		unsigned long int j=(cluster<<2)-(i<<9);	// errechnet das low byte von cluster in dem ausgelesenen sektor der fat
				
		if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden (fat_loadSector prüft ob schon gepuffert)			
			//printf("\n\rfat_setClust: %lu-> %lu",cluster,content);
			bytesOfSec=&sector[j];					// init des zeigers auf low byte	
			*bytesOfSec=(content&0x000000ff);		// low byte					
			bytesOfSec++;			
			*bytesOfSec=(content&0x0000ff00)>>8;	// 1. hi					
			bytesOfSec++;
			*bytesOfSec=(content&0x00ff0000)>>16;	// 2. hi
			bytesOfSec++;
			*bytesOfSec=(content&0xff000000)>>24;	// high byte	
			bufferDirty=1;									// zeigt an, dass im aktuellen sector geschrieben wurde
			return (0);
			}	
		}		


	return(1);									// neuladen des fat sektors, in dem oneCluster ist nötig !!		
}


//***************************************************************************************************************
// löscht cluster chain, beginnend ab dem startCluster.
// sucht cluster, setzt inhalt usw.. abschließend noch den cluster-chain ende markierten cluster löschen.
//***************************************************************************************************************
unsigned char fat_delClusterChain(unsigned long int startCluster){

  unsigned long int nextCluster=startCluster;		// tmp variable, wegen verketteter cluster..

  do{
	  startCluster=nextCluster;
	  nextCluster=fat_getNextCluster(startCluster);
	  fat_setCluster(startCluster,0x00000000);  

	//}while(nextCluster!=0x0ffffff8 && nextCluster!=0x0fffffff || (nextCluster!=0xffff && fat.fatType==16));
	}while(!((nextCluster==0xffffff8&&fat.fatType==32)||(nextCluster==0xfffffff&&fat.fatType==32)||(nextCluster==0xffff&&fat.fatType==16)));
	
	 fat_setCluster(startCluster,0x00000000);	 
	 if(bufferDirty==1)fat_writeSector(fat.currentSectorNr);
  return(0);
}


//************************************************************************************************<<***************
// Initialisiert die Fat(16/32) daten, wie 1. root directory sektor, 2. daten sektor, 3. fat sektor
//	siehe auch Fatgen103.pdf
// byt/sector, byt/cluster, anzahlDerFat, sector/fat ...
//*****************************************************************<**********************************************
unsigned char fat_loadFatData(unsigned long int sec){

 									  // offset,size			
	unsigned int  		rootEntCnt;		// 17,2				?!?		(fat16=512)
	unsigned int  		fatSz16;			// 22,2				sectors occupied by one fat16
	unsigned long int fatSz32;			// 36,4				sectors occupied by one fat32

	if(0==mmc_read_sector(sec,sector)){	// lesen von fat sector und bestimmen der wichtigen berreiche 
		
		fat.secPerClust=sector[13];		// fat.secPerClust, 13 only (power of 2)
		
		fat.bytsPerSec=sector[12];			// bytesPerSec, 12 high nibble 11 low nibble
		fat.bytsPerSec=fat.bytsPerSec<<8;
		fat.bytsPerSec+=sector[11];
		
		fat.fatSec=sector[15];				// resvdSecCnt, 15 high nibble 14 low nibble
		fat.fatSec=fat.fatSec<<8;
		fat.fatSec+=sector[14];

		rootEntCnt=sector[18];				// rootEntCnt, 18 high nibble 17 low nibble
		rootEntCnt=rootEntCnt<<8;
		rootEntCnt+=sector[17];
		
		fatSz16=sector[23];					// fatSz16, 23 high nibble 22 low nibble
		fatSz16=fatSz16<<8;
		fatSz16+=sector[22];
		
		
		fat.rootDir	 = (((rootEntCnt <<5) + fat.bytsPerSec) / fat.bytsPerSec)-1;		// ist 0 bei fat 32, sonst der root dir sektor	

		if(fat.rootDir==0){													// FAT32 spezifisch
			unsigned long int tmp;
			fatSz32=0;

			tmp=sector[39];				// hi byte					auslesen der fatSz32
			fatSz32 |=(tmp<<24);		
			tmp=sector[38];				// 1 lower byte
			fatSz32 |=(tmp<<16);	
			tmp=sector[37];				// 2 lower byte
			fatSz32 |=(tmp<<8);
			tmp=sector[36];				// lowest byte
			fatSz32 |=tmp;
								
			tmp=sector[47];				// hi byte					auslesen des Root Direktory (wird bei fat32 angegeben als clusterNr)
			fat.rootDir	 |=(tmp<<24);	
			tmp=sector[46];				// 1 lower byte
			fat.rootDir	 |=(tmp<<16);	
			tmp=sector[45];				// 2 lower byte
			fat.rootDir	 |=(tmp<<8);
			tmp=sector[44];				// lowest byte
			fat.rootDir	 |=tmp;														//root Dir cluster !! (da normal im daten bereich)				

			fat.dataDirSec = fat.fatSec + (fatSz32 * sector[16]);			// data sector (beginnt mit cluster 2)		
			fat.dir='0';																// dir auf 0==root dir, sonst 1. Cluster des dir											
			fat.dataDirSec-=2;														// zeigt auf 1. cluster			
			fat.fatType=32;															// fat typ					
			fat.dataDirSec+=sec;														// addiert den startsektor auf (umrechnung von absolut auf real)
			fat.fatSec+=sec;															// addiert den startsektor auf 	"
	
	
			}

		else{																	// FAT16	spezifisch
			fat.dataDirSec = fat.fatSec + (fatSz16 * sector[16]) + fat.rootDir;		// data sektor (beginnt mit cluster 2)
			fat.rootDir=fat.dataDirSec-fat.rootDir;											// root dir sektor, da nicht im datenbereich(cluster)
			fat.dataDirSec-=2;																		// zeigt auf 1. cluster		
			fat.dataDirSec+=sec;																		// addiert den startsektor auf (umrechnung von absolut auf real)
			fat.fatSec+=sec;																			// addiert den startsektor auf 	"
			fat.rootDir+=sec;																			// addiert den startsektor auf 	"
			fat.dir='0';																				// dir auf '0'==root dir, sonst 1.Cluster des dir
			fat.fatType=16;																			// fat typ			
			}		

		return (0);
		}


	return (1);			// sector nicht gelesen, fat nicht initialisiert!!









}


//************************************************************************************************<<***************
// int fat 
//************************************************************************************************<<***************
unsigned char fat_initfat(void){					
  
  unsigned long int secOfFirstPartition=0;						// ist 1. sektor der 1. partition aus dem MBR

	if(superfloppy==0){													// ist partitioniert
		if(0==mmc_read_sector(0,sector)){	
		
			unsigned long int tmp;			

			tmp=sector[457];				// hi byte					auslesen des 1. sectors der 1.partition
			secOfFirstPartition |=(tmp<<24);		
			tmp=sector[456];				// 1 lower byte
			secOfFirstPartition |=(tmp<<16);	
			tmp=sector[455];				// 2 lower byte
			secOfFirstPartition |=(tmp<<8);
			tmp=sector[454];				// lowest byte
			secOfFirstPartition |=tmp;
			
			return fat_loadFatData(secOfFirstPartition);			// läd fat daten aus dem 1. sektor der patition
			}

		}

	else {
		return fat_loadFatData(secOfFirstPartition);				// ist nicht partitioniert, läd fat daten aus sektor 0
		}

	return (1);
}
